/*-
 * Copyright (c) 2001-2002 Andrey Simonenko
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
  "@(#)$Id: kpf.c,v 1.1.2.6 2002/10/28 21:40:39 simon Exp $";
#endif /* !lint */

#include "kpf.h"

#ifdef WITH_PF

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <varargs.h>

#include "debug.h"

#define PFDEV	"/dev/pf"

u_quad_t	*kpf;
u_int		nkpf;

static const char	*kpfdev = PFDEV;
static int		kpffd = -1;

/*
 * Init Packet Filter support: open /dev/pf device.
 */
int
kpf_init(void)
{
	char	errmsg[] = "Packet Filter is not configured in the kernel or does not work prperly";

	syslog(LOG_INFO, "init Packet Filter support");
	if ( (kpffd = open(kpfdev, O_RDWR)) < 0) {
		syslog(LOG_ERR, "kpf_init: open(%s, O_RDWR): %m", kpfdev);
		syslog(LOG_ERR, "%s", errmsg);
		return -1;
	}
	if (debug_pf > 0)
		syslog(LOG_INFO, "kpf_init: %s opened, FD %d", kpfdev, kpffd);
	kpf = NULL;
	nkpf = 0;
	return 0;
}

/*
 * Remover Packet Filter support: close descriptor for opened /dev/pf.
 */
void
kpf_close(void)
{
	syslog(LOG_INFO, "remove Packet Filter support");
	if (kpffd != -1) {
		if (close(kpffd) < 0)
			syslog(LOG_ERR, "kpf_close: close(FD %d (%s)): %m",
			    kpffd, kpfdev);
		kpffd = -1;
	}
	free(kpf);
}

/*
 * Read the table of PF rules from the kernel.
 * Reallocate internal structures if needed.
 */
int
kpf_read_table(void)
{
	u_int		i;
	struct pfioc_rule	pfr;

	if (ioctl(kpffd, DIOCGETRULES, &pfr) < 0) {
		syslog(LOG_ERR, "kpf_read_table: ioctl(DIOCGETRULES): %m");
		return -1;
	}
	if (nkpf != pfr.nr) {
		if ( (kpf = realloc(kpf, pfr.nr * sizeof *kpf)) == NULL) {
			syslog(LOG_ERR, "kpf_read_table: realloc(%u bytes): %m",
			    pfr.nr * sizeof *kpf);
			return -1;
		}
		nkpf = pfr.nr;
		if (debug_pf > 0)
			syslog(LOG_INFO, "kpf_read_table: realloc, nkpf = %u", nkpf);
	}
	for (i = 0; i < nkpf; ++i) {
		pfr.nr = i;
		if (ioctl(kpffd, DIOCGETRULE, &pfr) < 0) {
			syslog(LOG_ERR, "kpf_read_table: ioctl(DIOCGETRULE, nr = %u): %m", i);
			return -1;
		}
		kpf[i] = pfr.rule.bytes;
	}
	if (debug_pf > 1) {
		u_int	i;
		int	len, total_len;
		char	one[10 + 1 + 1]; /* 10 - length of u_int, 1 - ' ', 1 - '\0' */
#define TXT_LEN 60
		char	txt[TXT_LEN + 1];

		syslog(LOG_INFO, "kpf_read_table: dump PF table (%u record%s):",
		    nkpf, nkpf > 1 ? "s" : "");
		for (i = 0, total_len = 0; i < nkpf; ++i) {
			len = snprintf(one, sizeof(one), "%u ", i);
			if (total_len + len > TXT_LEN) {
				syslog(LOG_INFO, "** %s", txt);
				total_len = 0;
			}
			strcpy(txt + total_len, one);
			total_len += len;
		}
		if (total_len != 0)
			syslog(LOG_INFO, "** %s", txt);
	}
	return 0;
}

#endif /* WITH_PF */
